@@ -1,12 +1,12 @@
-# (c) Alexander Solovyov, 2009-2010, under terms of the new BSD License
+# (c) Alexander Solovyov, 2009-2011, under terms of the new BSD License
 '''Command line arguments parser
 '''
 
-import sys, traceback, getopt, types, textwrap, inspect, os
+import sys, traceback, getopt, types, textwrap, inspect, os, copy
 from itertools import imap
 
 __all__ = ['command', 'dispatch']
-__version__ = '1.2'
+__version__ = '2.0'
 __author__ = 'Alexander Solovyov'
 __email__ = 'piranha@piranha.org.ua'
 
@@ -18,7 +18,8 @@ try:
 except locale.Error:
     ENCODING = 'UTF-8'
 
-def write(text, out=sys.stdout):
+def write(text, out=None):
+    out = out or sys.stdout
     if isinstance(text, unicode):
         return out.write(text.encode(ENCODING))
     out.write(text)
@@ -69,6 +70,7 @@ def command(options=None, usage=None, name=None, shortlist=False, hide=False):
 
         def help_func(name=None):
             return help_cmd(func, replace_name(usage_, sysname()), options_)
+        func.help = help_func
 
         @wraps(func)
         def inner(*args, **opts):
@@ -81,23 +83,23 @@ def command(options=None, usage=None, name=None, shortlist=False, hide=False):
 
             argv = opts.pop('argv', sys.argv[1:])
             if opts.pop('help', False):
-                return help_func()
+                return func.help()
 
             if args or opts:
                 # no catcher here because this is call from Python
                 return call_cmd_regular(func, options_)(*args, **opts)
 
             try:
-                opts, args = catcher(lambda: parse(argv, options_), help_func)
+                opts, args = catcher(lambda: parse(argv, options_), func.help)
             except Abort:
                 return -1
 
             if opts.pop('help', False):
-                return help_func()
+                return func.help()
 
             try:
                 return catcher(lambda: call_cmd(name_, func)(*args, **opts),
-                               help_func)
+                               func.help)
             except Abort:
                 return -1
 
@@ -181,8 +183,8 @@ def help_(cmdtable, globalopts):
                 if shortlist and not cmd.startswith('^'):
                     continue  # short help contains only marked commands
                 cmd = cmd.lstrip('^~')
-                doc = info[0].__doc__.strip() or '(no help text available)'
-                hlp[cmd] = doc.splitlines()[0].rstrip()
+                doc = info[0].__doc__ or '(no help text available)'
+                hlp[cmd] = doc.strip().splitlines()[0].rstrip()
 
             hlplist = sorted(hlp)
             maxlen = max(map(len, hlplist))
@@ -191,11 +193,7 @@ def help_(cmdtable, globalopts):
             write('\ncommands:\n\n')
             for cmd in hlplist:
                 doc = hlp[cmd]
-                if False:  # verbose?
-                    write(' %s:\n     %s\n' % (cmd.replace('|', ', '), doc))
-                else:
-                    write(' %-*s  %s\n' % (maxlen, cmd.split('|', 1)[0],
-                                              doc))
+                write(' %-*s  %s\n' % (maxlen, cmd.split('|', 1)[0], doc))
 
         if not cmdtable:
             return err('No commands specified!\n')
@@ -242,7 +240,6 @@ def help_cmd(func, usage, options):
      -p --port       port to listen on (default: 8000)
      -d --daemonize  daemonize process
         --pid-file   name of file to write process ID to
-    <BLANKLINE>
     '''
     write(usage + '\n')
     doc = func.__doc__ or '(no help text available)'
@@ -251,6 +248,8 @@ def help_cmd(func, usage, options):
         write(''.join(help_options(options)))
 
 def help_options(options):
+    '''Generator for help on options
+    '''
     yield 'options:\n\n'
     output = []
     for o in options:
@@ -309,8 +308,8 @@ def parse(args, options):
         defmap[pyname] = default
 
         # copy defaults to state
-        if isinstance(default, list):
-            state[pyname] = default[:]
+        if isinstance(default, (list, dict)):
+            state[pyname] = copy.copy(default)
         elif hasattr(default, '__call__'):
             funlist.append(pyname)
             state[pyname] = None
@@ -319,8 +318,10 @@ def parse(args, options):
 
         # getopt wants indication that it takes a parameter
         if not (default is None or default is True or default is False):
-            if short: short += ':'
-            if name: name += '='
+            if short:
+                short += ':'
+            if name:
+                name += '='
         if short:
             shortlist += short
         if name:
@@ -335,8 +336,15 @@ def parse(args, options):
         if t is types.FunctionType:
             del funlist[funlist.index(name)]
             state[name] = defmap[name](val)
-        elif t is types.ListType:
+        elif t is list:
             state[name].append(val)
+        elif t is dict:
+            try:
+                k, v = val.split('=')
+            except ValueError:
+                raise ParseError(name, "wrong definition: '%s' "
+                                 "(should be in format KEY=VALUE)" % val)
+            state[name][k] = v
         elif t in (types.NoneType, types.BooleanType):
             state[name] = not defmap[name]
         else:
@@ -470,7 +478,7 @@ def catcher(target, help_func):
             (e.args[0], ' '.join(e.args[1])))
         raise Abort()
     except ParseError, e:
-        err('%s: %s\n' % (e.args[0], e.args[1]))
+        err('%s: %s\n\n' % (e.args[0], e.args[1].strip()))
         help_func(e.args[0])
         raise Abort()
     except getopt.GetoptError, e:
@@ -496,6 +504,8 @@ def call_cmd(name, func, depth=1):
     return inner
 
 def call_cmd_regular(func, opts):
+    '''Wrapper for command for handling function calls from Python
+    '''
     def inner(*args, **kwargs):
         funcargs, _, varkw, defaults = inspect.getargspec(func)
         if len(args) > len(funcargs):
@@ -644,3 +654,7 @@ class ParseError(OpsterError):
 
 class Abort(OpsterError):
     'Processing error, abort execution'
+
+if __name__ == '__main__':
+    import doctest
+    doctest.testmod()
